﻿using BuildSoft.OscCore;

namespace VRC_Console;

public class OSCCollar
{

    public static void Main(string[] args)
    {
        OSCCollar collar = new OSCCollar();
    }
    
    private OscServer Server { get; init; }
    private OscClient Client { get; init; }
    private float _collarAngle, _collarStretch;
    private double _leftOrRight = 1;
    private double _verticalMovement = 0;
    private double _horizontalMovement = 0;
    private bool _allowMoving = false;
    private uint _nilSent = 0;
    private DateTime _lastNilMessageSent = DateTime.UnixEpoch;
    private DateTime _lastConsoleOutput = DateTime.UnixEpoch;
    private static readonly TimeSpan ConsoleUpdateInterval = TimeSpan.FromMilliseconds(50);
    private static readonly TimeSpan UpdateFieldsTimeout = TimeSpan.FromMilliseconds(1);
    private static readonly TimeSpan NilMessageMaxTimeout = TimeSpan.FromMilliseconds(400);

    private OSCCollar()
    {
        this.Server = new OscServer(9001);
        this.Client = new OscClient("127.0.0.1", 9000);
        this._collarAngle = 0;
        this._collarStretch = 0;
        Server.TryAddMethod("/avatar/parameters/Collar_Angle", CollarAngleHandle);
        Server.TryAddMethod("/avatar/parameters/CollarRight", CollarLeftRightHandle);
        Server.TryAddMethod("/avatar/parameters/Collar_Allow_Moving", AllowMovingHandle);
        Server.TryAddMethodPair("/avatar/parameters/Collar_Stretch", CollarStretchHandle, CalculateMovement);
        this.Server.Start();

        Thread runningThread = new Thread(RunningThread);
        runningThread.Start();
    }

    private void PrintOutput(double verticalMovement, double horizontalMovement)
    {
        Console.Clear();
        Console.WriteLine($"OSC Collar - Status: {(_allowMoving ? "enabled" : "disabled")}");
        Console.WriteLine("===============================================================");
        Console.WriteLine($"Angle:...............{_collarAngle:0.000000}");
        Console.WriteLine($"Left/Right:..........{(_leftOrRight < 0 ? "left" : "right")}");
        Console.WriteLine($"Stretch:.............{_collarStretch:0.000000}");
        Console.WriteLine($"Vertical Movement:...{verticalMovement:0.000000}");
        Console.WriteLine($"Horizontal Movement:.{horizontalMovement:0.000000}");
        Console.WriteLine("");
        Console.WriteLine("");
        Console.WriteLine("");
        Console.WriteLine("");
        Console.WriteLine("");
        Console.WriteLine("");
        Console.WriteLine("");
        Console.WriteLine($"Sending /input nil {_nilSent}");
    }

    private void AllowMovingHandle(OscMessageValues messageValues)
    {
        this._allowMoving = messageValues.ReadBooleanElement(0);
    }

    private void CollarAngleHandle(OscMessageValues messageValues)
    {
        this._collarAngle = messageValues.ReadFloatElement(0);
    }

    private void CollarLeftRightHandle(OscMessageValues messageValues)
    {
        this._leftOrRight = messageValues.ReadBooleanElement(0) ? 1 : -1;
    }

    private void CollarStretchHandle(OscMessageValues messageValues)
    {
        this._collarStretch = messageValues.ReadFloatElement(0);
    }

    private const float PowerAdjust = 0.25f;
    private void CalculateMovement()
    {
        float degrees = _collarAngle * 180;
        double radians = degrees * Math.PI / 180;
        _verticalMovement = Math.Cos(radians) * this._collarStretch * PowerAdjust;
        _horizontalMovement = Math.Sin(radians) * this._collarStretch * PowerAdjust * _leftOrRight;
    }

    private void RunningThread()
    {
        while (true)
        {
            this.Server.Update();
            if (_lastNilMessageSent.Add(NilMessageMaxTimeout) < DateTime.Now)
            {
                this.Client.Send("/input/Vertical", 0f);
                this.Client.Send("/input/Horizontal", 0f);
                this._lastNilMessageSent = DateTime.Now;
                _nilSent++;
            }

            if (_allowMoving)
            {
                Thread.Sleep(1);
                this.Client.Send("/input/Vertical", Convert.ToSingle(_verticalMovement));
                this.Client.Send("/input/Horizontal", Convert.ToSingle(_horizontalMovement));
            }

            if (_lastConsoleOutput.Add(ConsoleUpdateInterval) < DateTime.Now)
            {
                PrintOutput(_verticalMovement, _horizontalMovement);
                _lastConsoleOutput = DateTime.Now;
            }
            
            Thread.Sleep(UpdateFieldsTimeout);
        }
    }
}